# -*- coding: ISO-8859-1 -*-
# Copyright (C) 2002-2003 Juan David Ibez Palomar <jdavid@itaapy.com>
#               2003 Itaapy <contact@itaapy.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA


__revision__ = "$Id: index.py,v 1.11 2003/08/20 09:52:33 jdavid Exp $"


# Import Python modules
import sets
from types import ListType, TupleType

# Import itools modules
from itools.upgrade import Upgradeable

# Import ZODB modules
try:
    import ZODB
    from Persistence import Persistent, PersistentMapping
except ImportError:
    # To run the unit tests
    class Persistent:
        pass

    from UserDict import UserDict as PersistentMapping



class Index(Upgradeable, Persistent):
    """
    Instances of this class let to index and search objects. The indexed
    objects are identified by an id (oid).

    There are two internal data structures:

      - forward: keeps a mapping from value to oid

      - reverse: keeps a mapping from oid to value
    """


    __version__ = '20030804'


    def __init__(self):
        Upgradeable.__init__(self)

        self.forward = PersistentMapping()
        self.reverse = PersistentMapping()


    #########################################################################
    # API
    #########################################################################
    def add(self, oid, *args):
        """
        Indexes an object by the given value.
        """
        # Sets the forward index
        for value in args:
            if not self.forward.has_key(value):
                self.forward[value] = sets.ImmutableSet()

            self.forward[value] |= sets.ImmutableSet([oid])

        # Sets the reverse index
        if not self.reverse.has_key(oid):
            self.reverse[oid] = sets.ImmutableSet()

        self.reverse[oid] |= sets.ImmutableSet(args)



    def remove(self, oid):
        """
        Unindex the given object.
        """
        if self.reverse.has_key(oid):
            # Remove from the forward index
            for value in self.reverse[oid]:
                self.forward[value] -= sets.ImmutableSet([oid])

                # Remove value
                if len(self.forward[value]) == 0:
                    del self.forward[value]

            # Remove from the reverse index
            del self.reverse[oid]



class ValueIndex(Index):
    """ """

    def _is(self, value):
        return self.forward.get(value, sets.ImmutableSet())


    def _in(self, values):
        set = sets.Set()
        for value in values:
            set = set | self._is(value)
        return set


    def _starts(self, value):
        set = sets.Set()
        for k in self.forward.keys():
            if k is not None and k.startswith(value):
                set = set | self.forward[k]
        return set


class TextIndex(ValueIndex):
    def add(self, oid, *args):
        for arg in args:
            values = arg.split()
            apply(Index.add, (self, oid) + tuple(values))



class ExprIndex(Index):
    """ """

    def search(self, values=()):
        """
        Returns a list with the oids of all the matching objects.
        An object matchs if it has *any* of the given values.
        """
        if type(values) is ListType:
            values = tuple(values)
        elif type(values) is not TupleType:
            values = (values,)

        oids = sets.Set()
        for expr in self.forward.keys():
            f = eval(expr)  # XXX This could be optimized
            if apply(f, values):
                oids.union_update(self.forward[expr])

        return oids
