#!/usr/bin/env python
# Copyright (C) 2003 Itaapy <contact@itaapy.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA


# Import from Python
import codecs
from HTMLParser import HTMLParser
try:
    # Python 2.3
    import optparse
except ImportError:
    try:
        import optik as optparse
    except ImportError:
        optparse = None
import os
import re
from StringIO import StringIO
import sys
import tempfile
import time
import types
from xml.parsers import expat



############################################################################
# PO
############################################################################
# Exceptions
class ParseError(Exception): pass


class PO:
    """
    XXX Maybe the implementation could be changed to use the gettext tools.
    """

    def __init__(self, data):
        if type(data) is types.FileType:
            data = data.readlines()
        elif type(data) in types.StringTypes:
            data = data.split('\n')

        # The regular expressions
        com = re.compile('^#.*')
        msgid = re.compile(r'^ *msgid *"(.*[^\\]*)"')
        msgstr = re.compile(r'^ *msgstr *"(.*[^\\]*)"')
        re_str = re.compile(r'^ *"(.*[^\\])"')
        blank = re.compile(r'^\s*$')
        quote_esc = re.compile(r'\\"')

        messages = {}
        msgids = []
        pointer = 0
        state = 0
        COM, MSGID, MSGSTR = [], [], []
        while pointer < len(data):
            line = data[pointer]
            # print 'STATE:', state, 'LANGUAGE', language
            # print 'LINE:', data[pointer].strip()
            if state == 0:
                COM, MSGID, MSGSTR = [], [], []
                if com.match(line):
                    COM.append(line.strip())
                    state = 1
                    pointer = pointer + 1
                elif msgid.match(line):
                    line = msgid.match(line).group(1)
                    MSGID.append(self.unescape(line))
                    state = 2
                    pointer = pointer + 1
                elif blank.match(line):
                    pointer = pointer + 1
                else:
                    raise ParseError, 'state 0, line %d\n' % (pointer + 1)
            elif state == 1:
                if com.match(line):
                    COM.append(line.strip())
                    state = 1
                    pointer = pointer + 1
                elif msgid.match(line):
                    line = msgid.match(line).group(1)
                    MSGID.append(self.unescape(line))
                    state = 2
                    pointer = pointer + 1
                elif blank.match(line):
                    pointer = pointer + 1
                else:
                    raise ParseError, 'state 1, line %d\n' % (pointer + 1)
            elif state == 2:
                if com.match(line):
                    COM.append(line.strip())
                    state = 2
                    pointer = pointer + 1
                elif re_str.match(line):
                    line = re_str.match(line).group(1)
                    MSGID.append(self.unescape(line))
                    state = 2
                    pointer = pointer + 1
                elif msgstr.match(line):
                    line = msgstr.match(line).group(1)
                    MSGSTR.append(self.unescape(line))
                    state = 3
                    pointer = pointer + 1
                elif blank.match(line):
                    pointer = pointer + 1
                else:
                    raise ParseError, 'state 2, line %d\n' % (pointer + 1)
            elif state == 3:
                if com.match(line) or msgid.match(line):
                    # print "\nEn", language, "detected", MSGID
                    MSGID = tuple(MSGID)
                    if messages.has_key(MSGID):
                        raise ParseError, \
                              'msgid more than once, line %d\n' % (pointer + 1)
                    messages[MSGID] = (COM, MSGSTR)
                    msgids.append(MSGID)
                    state = 0
                elif re_str.match(line):
                    line = re_str.match(line).group(1)
                    MSGSTR.append(self.unescape(line))
                    state = 3
                    pointer = pointer + 1
                elif blank.match(line):
                    pointer = pointer + 1
                else:
                    raise ParseError, 'state 3, line %d\n' % (pointer + 1)

        # Detect the encoding
        if messages.has_key(('',)):
            for x in messages[('',)][1]:
                if x.startswith('Content-Type: '):
                    break
            x = x.strip()
            encoding = x.split('charset=')[-1]
        else:
            encoding = 'ascii'

        try:
            codecs.lookup(encoding)
        except LookupError:
            # Default encoding
            # XXX A warning should be issued
            encoding = 'utf8'

        # Change (only) the translations to unicode strings
        for k, v in messages.items():
            aux = [ type(x) is types.StringType and unicode(x, encoding) or x
                    for x in v[1] ]
            messages[k] = (v[0], aux)

        # the last also goes in
        MSGID = tuple(MSGID)
        if MSGID:
            if messages.has_key(MSGID):
                raise ParseError, \
                      'msgid more than once, line %d\n' % (pointer + 1)
            messages[MSGID] = (COM, MSGSTR)
            msgids.append(MSGID)

        #
        self.encoding = encoding
        self.messages = messages
        self.msgids = msgids


    def escape(self, s):
        return s.replace('\n', '\\n')


    def unescape(self, s):
        s2 = ''
        state = 0
        for x in s:
            if state == 0:
                if x == '\\':
                    state = 1
                else:
                    s2 = s2 + x
            else:
                s2 = s2 + eval("'\%s'" % x)
                state = 0

        return s2


    def write(self):
        output = []
        for msgid in self.msgids:
            comments, msgstr = self.messages[msgid]

            for line in comments:
                output.append(line)

            output.append('msgid "%s"' % self.escape(msgid[0]))
            for line in msgid[1:]:
                output.append('"%s"' % self.escape(line))

            output.append('msgstr "%s"' % self.escape(msgstr[0]))
            for line in msgstr[1:]:
                output.append('"%s"' % self.escape(line))

            output.append('')

        return '\n'.join(output)





############################################################################
# XML
############################################################################
class Raw:
    """
    Represents raw text in an XML element.
    """

    def __init__(self, data):
        self.data = data


    def get_messages(self, elements, attributes):
        return []


    def translate(self, translations, elements, attributes, html):
        return self.data


    def toxml(self, html=0):
        return self.data


    tohtml = toxml


class Element:
    """
    Represents an XML (or HTML) element.
    """

    def __init__(self, name, attrs, dummy):
        self.name = name
        self.attrs = attrs

        # Dummy nodes are those that we are not going to process
        self.dummy = dummy

        self.children = []


    def get_messages(self, elements, attributes):
        messages = []

        # Attributes
        for attribute_name, condition in attributes.get(self.name, []):
            # Check condition
            if not condition or self.attrs.get(condition[0]) == condition[1]:
                message = self.attrs.get(attribute_name)
                if message and message not in messages:
                    messages.append(message)

        # Content
        if self.name in elements:
            message = self.content().strip()
            message = ' '.join([ x.strip() for x in message.split('\n') ])
            message = message.replace('"', '\\"')

            # Store the message if it is not already there
            if message and message not in messages:
                messages.append(message)

        # Children
        for child in self.children:
            for message in child.get_messages(elements, attributes):
                if message not in messages:
                    messages.append(message)

        return messages


    def translate(self, translations, elements, attributes, html):
        attrs = self.attrs.copy()
        for attribute_name, condition in attributes.get(self.name, []):
            # Check condition
            if not condition or self.attrs.get(condition[0]) == condition[1]:
                message = self.attrs.get(attribute_name)
                if message:
                    translation = translations.get(message)
                    if translation:
                        attrs[attribute_name] = translation

        s = '<' + self.name
        for key, value in attrs.items():
            s += ' %s="%s"' % (key, value)
        s += '>'

        if html and self.name in ['br', 'img', 'input']:
            return s

        if self.name in elements:
            content = self.content()
            message = content.strip()
            message = ' '.join([ x.strip() for x in message.split('\n') ])
            if message:
                if type(message) is not types.UnicodeType:
                    message = unicode(message, 'utf-8')
                message = translations.get(message) or content
                s = s + message
            else:
                s = s + content
        else:
            for child in self.children:
                s = s \
                    + child.translate(translations, elements, attributes, html)

        return s + '</' + self.name + '>'


    def content(self, html=0):
        """
        Returns the elements content as text.
        """
        s = ''
        for child in self.children:
            s += child.toxml(html)

        return s


    def toxml(self, html=0):
        """
        Outputs as XML.

        If the html parameter is true output will be as html.
        """
        # Open tag
        s = '<' + self.name
        for key, value in self.attrs.items():
            s += ' %s="%s"' % (key, value)
        s += '>'

        if html and self.name in ['br', 'img', 'input']:
            return s
        else:
            return s + self.content(html) + '</' + self.name + '>'


class Document:
    def __init__(self):
        self.children = []


    def get_messages(self, elements, attributes):
        messages = []
        for child in self.children:
            for message in child.get_messages(elements, attributes):
                if message not in messages:
                    messages.append(message)

        return messages


    def translate(self, translations, elements, attributes, html):
        s = ''
        for child in self.children:
            s = s + child.translate(translations, elements, attributes, html)
        return s


    def toxml(self, html=0):
        s = ''
        for child in self.children:
            s = s + child.toxml(html)
        return s



class XML:
    """
    Generic XML parser. It extracts the translatable strings from an XML
    document and returns them.
    """

    def __init__(self, data, encoding='utf-8'):
        """
        Parses the given xml and builds a data structure where each node
        is an element.

        This is a mixin class intended to be specialized for any particular
        XML language (xhtml for example). Derived classes must provide the
        variables:

          - elements
            List of element names whose content is translatable.

            Example: ['p', ...]

          - attributes
            Dictionary whose keys are element names, the values are the
            list of attributes of this elements that are translatable.

            Example: {'img': ['alt', ...]}
        """
        # Preprocess data
        data = self.preprocess_data(data, encoding)
        file = StringIO(data)

        # Add parsing semantics
        p = expat.ParserCreate()
        p.StartElementHandler = self.start_element
        p.EndElementHandler = self.end_element
        p.DefaultHandler = self.default

        # Parse file
        document = Document()
        self.stack = [document]
        p.ParseFile(file)
        del self.stack

        self.document = document


    def preprocess_data(self, data, encoding='utf-8'):
        """
        The class constructor takes the parameter 'data' as input, it can
        be a normal string, an unicode string or a file. This method is
        called before the data is processed, it gets the data and returns
        it as a normal string.
        """
        if type(data) is types.UnicodeType:
            # It is a unicode string, encode it
            data = data.encode(encoding)
        elif type(data) is not types.StringType:
            # It is a file, read it
            data = data.read()

        # Escape internal entities
        return data.replace('&', '&amp;')


    def start_element(self, name, attrs):
        dummy = 1
        if name in self.elements:
            dummy = 0
        else:
            for attribute_name, condition in self.attributes.get(name, []):
                # Check condition
                if not condition or attrs.get(condition[0]) == condition[1]:
                    dummy = 0

        element = Element(name, attrs, dummy)
        self.stack.append(element)


    def end_element(self, name, html=0):
        """
        The parameter index lets to implement non XML parsers, for example
        an HTML parser.
        """
        element = self.stack.pop(-1)

        if element.dummy and not [ x for x in element.children
                                   if isinstance(x, Element) ]:
            self.default(element.toxml(html))
        else:
            self.stack[-1].children.append(element)


    def default(self, data):
        # Unscape internal entities
        data = data.replace('&amp;', '&')

        children = element = self.stack[-1].children
        if children and isinstance(children[-1], Raw):
            children[-1].data += data
        else:
            children.append(Raw(data))


    ########################################################################
    # Get messages
    ########################################################################
    def get_messages(self):
        """ """
        return self.document.get_messages(self.elements, self.attributes)


    ########################################################################
    # Translate
    ########################################################################
    def translate(self, po_file, html=0):
        """ """
        # Parse PO file
        translations = {}
        po = PO(po_file)
        for k, v in po.messages.items():
            translations[''.join(k)] = ''.join(v[1])

        return self.document.translate(translations, self.elements,
                                       self.attributes, html)



############################################################################
# XHTML
class XHTML(XML):
    elements = ['p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'option', 'th']
    attributes = {'img': [('alt', None)],
                  'input': [('value', ('type', 'submit'))]}



############################################################################
# HTML
############################################################################
class HTML(XHTML, HTMLParser):
    """ """

    def __init__(self, data, encoding='utf-8'):
        # Call base classes constructors
        HTMLParser.__init__(self)

        # Preprocess data
        data = self.preprocess_data(data, encoding)

        # Parse data
        document = Document()
        self.stack = [document]
        self.feed(data)
        self.close()
        del self.stack

        self.document = document


    def handle_starttag(self, name, attrs):
        # Change attrs from a list to a dictionary, as needed by XHTML
        attributes = {}
        for k, v in attrs:
            attributes[k] = v

        XHTML.start_element(self, name, attributes)

        # Close the tag if needed
        if name in ['br', 'img', 'input']:
            XHTML.end_element(self, name, html=1)


    def handle_endtag(self, name):
        # Ignore end tags without an start
        if name == getattr(self.stack[-1], 'name', None):
            XHTML.end_element(self, name, html=1)


    handle_data = XHTML.default


    def handle_entityref(self, name):
        self.handle_data('&%s;' % name)


    def close(self):
        while len(self.stack) > 1:
            XHTML.end_element(self, self.stack[-1].name)
        HTMLParser.close(self)


    def translate(self, po_file):
        return XHTML.translate(self, po_file, 1)



############################################################################
# Return PO file
############################################################################
def run():
    # Parse the options
    parser = optparse.OptionParser()
    parser.add_option('-o', '--output',
                      dest='output',
                      help="The output will be sent to the given file,"
                           " instead of to the standard output.")
    parser.add_option('--pot',
                      action="store_const", const=0, dest='action',
                      help="Extracts the translatable messages and outputs"
                           " the POT file")
    parser.add_option('--po',
                      action="store_const", const=1, dest='action',
                      help="Updates the po file with the messages from the"
                           " POT file")
    parser.add_option('--xhtml',
                      action="store_const", const=2, dest='action',
                      help="Outputs a new XHTML file from the original source"
                           " and a PO file")
    options, args = parser.parse_args()

    # Action
    if options.action == 0:
        # Create one auxiliar PO file for each input file
        tmp_files = []
        for source_file in args[1:]:
            # Create the temp file and add it to the list
            tmp_file = tempfile.mktemp('.po')
            tmp_files.append(tmp_file)

            # Write the file's content
            if source_file.endswith('.py'):
                os.system('xgettext --omit-header --keyword=ugettext --keyword=N_ --output=%s %s' % (tmp_file, source_file))
            else:
                # XXX Suppose it is HTML
                file = open(tmp_file, 'w')
                xhtml = XHTML(open(source_file))
                for message in xhtml.get_messages():
                    file.write('msgid "%s"\n' % message)
                    file.write('msgstr ""\n')
                    file.write('\n')
                file.close()

        # Merge all the PO files
        filename = tempfile.mktemp('.po')
        command = 'msgcat --output-file=%s %s' % (filename,' '.join(tmp_files))
        os.system(command)
        tmp_files.append(filename)

        # Get the output
        output = os.popen('msgmerge %s %s' % (args[0], filename)).read()

        # Remove all the used files
        for file in tmp_files:
            os.remove(file)
    elif options.action == 1:
        pot = args[0]
        po = args[1]
        if os.path.exists(po):
            # a .po file already exist, merge it with locale.pot
            output = os.popen('msgmerge %s %s' % (po, pot)).read()
        else:
            # po doesn't exist, just copy locale.pot
            output = open(pot).read()
    elif options.action == 2:
        xhtml = XHTML(open(args[0]))
        output = xhtml.translate(open(args[1]))
    else:
        pass

    # Encode to UTF8 if needed
    if type(output) is types.UnicodeType:
        output = output.encode('utf8')

    if options.output is None:
        print output
    else:
        open(options.output, 'w').write(output)



if __name__ == '__main__':
    run()
