# -*- coding: ISO-8859-1 -*-
# Copyright (C) 2002-2003 Juan David Ibez Palomar <jdavid@itaapy.com>
#               2003 Itaapy <contact@itaapy.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA



__revision__ = "$Id: stl.py,v 1.19 2003/09/18 17:11:56 jdavid Exp $"


"""
STL stands for Simple Template Language, as it is the simplest template
language I could imagine.

It uses an xml namespace..
"""



# Python
from StringIO import StringIO
from types import DictType, StringType, UnicodeType
from UserList import UserList
from xml.parsers import expat


########################################################################
# Evaluate STL expressions
########################################################################

# STL exceptions
class LexicalError(Exception): pass
class SyntacticError(Exception): pass
class LookupError(Exception): pass


# Expressions tokens
TID, TSLASH, TOPEN, TCLOSE, TEOF, TCONTENT, TATTRIBUTES, TREPEAT = range(8)

keywords = {'content': TCONTENT, 'attributes': TATTRIBUTES, 'repeat': TREPEAT}


class Expression:
    """
    Parses and evaluates stl expressions.
    
    Examples of allowed expressions:
    
      a
      a(content)
      a(attributes/alt)
      a(literal)
      a/b/c
      a/b/c(content)
      ...
    """

    def __init__(self, expression, node=None):
        self.expression = expression

        # Initialize semantic structures
        self.path = ()
        self.parameters = ()
        self.repeat = False

        # Parsing
        self.index = 0
        self.node = node
        self.parse()
        del self.index
        del self.node


    ###################################################################
    # Lexical analysis
    ###################################################################
    def get_token(self):
        token = TEOF
        lexeme = None
        state = 0

        while self.index < len(self.expression):
            c = self.expression[self.index]
            if state == 0:
                self.index += 1
                if c.isalnum() or c in ('_', '.'):
                    lexeme = c
                    state = 1
                elif c == '/':
                    token = TSLASH
                    break
                elif c == '(':
                    token = TOPEN
                    break
                elif c == ')':
                    token = TCLOSE
                    break
                else:
                    raise LexicalError, c
            elif state == 1:
                if c.isalnum() or c in ('_', '.'):
                    lexeme += c
                    self.index += 1
                else:
                    break

        if lexeme is not None:
            token = keywords.get(lexeme, TID)

        return token, lexeme


    ###################################################################
    # Syntactic and semantic analysis. Grammar:
    #
    #   parse = TID parser1
    #           | TREPEAT TSLASH TID TSLASH TID
    #   parser1 = TEOF
    #             | TSLASH parse
    #             | TOPEN parser2
    #   parser2 = TCONTENT parser3
    #             | TATTRIBUTES TSLASH TID parser3
    #             | TID parser3
    #   parser3 = TCLOSE TEOF
    ###################################################################
    def parse(self):
        token, lexeme = self.get_token()
        if token == TID:
            self.path = self.path + (lexeme,)
            self.parser1()
            return
        elif token == TREPEAT:
            self.repeat = True
            token, lexeme = self.get_token()
            if token == TSLASH:
                token, lexeme = self.get_token()
                if token == TID:
                    self.path = self.path + (lexeme,)
                    token, lexeme = self.get_token()
                    if token == TSLASH:
                        token, lexeme = self.get_token()
                        if token == TID:
                            self.path = self.path + (lexeme,)
                            return

        raise SyntacticError, 'id expected'


    def parser1(self):
        token, lexeme = self.get_token()
        if token == TEOF:
            return
        elif token == TSLASH:
            self.parse()
            return
        elif token == TOPEN:
            self.parser2()
            return

        raise SyntacticError, 'unexpected token'


    def parser2(self):
        token, lexeme = self.get_token()
        if token == TCONTENT:
            value = self.node.toxml()
            self.parameters = (value,)
            self.parser3()
            return
        elif token == TATTRIBUTES:
            token, lexeme = self.get_token()
            if token == TSLASH:
                token, lexeme = self.get_token()
                if token == TID:
                    value = self.node.attrs[lexeme]
                    self.parameters = (value,)
                    self.parser3()
                    return
        elif token == TID:
            self.parameters = (lexeme,)
            self.parser3()
            return

        raise SyntacticError, 'unexpected token'


    def parser3(self):
        token, lexeme = self.get_token()
        if token == TCLOSE:
            token, lexeme = self.get_token()
            if token == TEOF:
                return

        raise SyntacticError, 'unexpected token'


    ###################################################################
    # Semantic process.
    ###################################################################
    def evaluate(self, stack, repeat):
        if self.repeat:
            stack = repeat

        # Traverse
        x = stack.lookup(self.path[0])
        for name in self.path[1:]:
            x = lookup(x, name)

        # Call
        if self.parameters:
            return apply(x, self.parameters)
        else:
            return self.end(x)


    def end(self, value):
        """
        Call the object at the end of the traversal to get the value,
        if it is callable.
        """
        if callable(value):
            try:
                value = value()
            except AttributeError, error_value:
                # XXX "callable" could return true even if the object is not
                # callable (see Python's documentation).
                #
                # This happens, for example, in the context of Zope, maybe
                # because of extension classes and acquisition. So we catch
                # the AttributeError exception, we should test also for the
                # exception value to be "__call__". This is dangereous
                # because we could hide real errors. Further exploration
                # needed..
                pass

        return value




###########################################################################
# Namespace
###########################################################################
def lookup(namespace, name):
    """
    Looks for a variable in a namespace (an instance, a mapping, etc..)
    """
    if isinstance(namespace, DictType):
        if name in namespace:
            return namespace[name]

    try:
        value = getattr(namespace, name)
    except AttributeError:
        # XXX Maybe we shouldn't try this one
        try:
            value = namespace[name]
        except KeyError:
            raise LookupError, name

    return value



class NamespaceStack(UserList):
    """
    This class represents a namespace stack as used by STL. A variable
    is looked up in the stack from the top to the bottom until found.
    """

    def lookup(self, name):
        stack = self[:]
        stack.reverse()
        for namespace in stack:
            try:
                return lookup(namespace, name)
            except LookupError:
                pass

        raise LookupError, name


###########################################################################
# The tree
###########################################################################


# List of Boolean attributes in HTML that should be rendered in
# minimized form (e.g. <img ismap> rather than <img ismap="">)
# From http://www.w3.org/TR/xhtml1/#guidelines (C.10)
boolean_html_attributes = ['compact', 'nowrap', 'ismap', 'declare', 'noshade',
                           'checked', 'disabled', 'readonly', 'multiple',
                           'selected', 'noresize', 'defer']


class Raw:
    def __init__(self, data):
        self.data = data


    def toxml(self):
        return self.data


    def process(self, stack, repeat):
        return self.data


class Element:
    def __init__(self, name, attrs, preprocess=1):
        self.name = name
        self.attrs = attrs
        self.children = []

        # Actions
        self.srepeat = None
        self.sif = None
        self.sifnot = None
        self.sattributes = None
        self.scontent = None

        # Preprocess..
        if preprocess:
            # Sets prefix
            if self.name == 'stl:block':
                prefix = ''
            else:
                prefix = 'stl:'

            # stl:repeat
            name = prefix + 'repeat'
            expr = self.attrs.get(name, None)
            if expr is not None:
                # Remove stl sttribute
                del self.attrs[name]

                # Get the name and values
                name, expr = expr.split(' ', 1)
                self.srepeat = (name, Expression(expr))

            # stl:if
            name = prefix + 'if'
            expr = self.attrs.get(name, None)
            if expr is not None:
                del self.attrs[name]
                self.sif = Expression(expr)

            # stl:ifnot
            name = prefix + 'ifnot'
            expr = self.attrs.get(name, None)
            if expr is not None:
                del self.attrs[name]
                self.sifnot = Expression(expr)

            # stl:attributes
            name = prefix + 'attributes'
            expr = self.attrs.get(name, None)
            if expr is not None:
                del self.attrs[name]

                attributes = []
                for x in expr.split(';'):
                    name, expr = x.strip().split(' ', 1)
                    attributes.append((name, Expression(expr, self)))

                self.sattributes = tuple(attributes)

            # stl:content
            name = prefix + 'content'
            expr = self.attrs.get(name, None)
            if expr is not None:
                del self.attrs[name]
                self.scontent = Expression(expr, self)


    def process(self, stack, repeat):
        """
        Process a node.

          stl:content - replaces the content by ..

        XXX Finish docstring.
        """
        s = ''

        # Repeat
        if self.srepeat is not None:
            name, expression = self.srepeat

            i = 0
            values = expression.evaluate(stack, repeat)
            nvalues = len(values)
            for value in values:
                # Create the new stack
                newstack = stack[:]
                newstack.append({name: value})

                newrepeat = repeat[:]
                value = {'index': i,
                         'start': i == 0,
                         'end': i == nvalues - 1}
                newrepeat.append({name: value})

                # Process and append the clone
                s = s + self.process1(newstack, newrepeat)

                # Increment counter
                i = i + 1

            return s

        return s + self.process1(stack, repeat)


    def process1(self, stack, repeat):
        """
        Process stl:if, stl:ifnot, stl:attributes and stl:content.
        """
        # Remove the element if the given expression evaluates to false
        if self.sif is not None:
            expression = self.sif
            if not expression.evaluate(stack, repeat):
                return ''

        # Remove the element if the given expression evaluates to true
        if self.sifnot is not None:
            expression = self.sifnot
            if expression.evaluate(stack, repeat):
                return ''

        # Print tag name
        head = '<' + self.name

        # Process attributes
        attrs = self.attrs.copy()
        if self.sattributes is not None:
            for name, expression in self.sattributes:
                value = expression.evaluate(stack, repeat)
                # XXX Do it only if it is an HTML document.
                if 1 and name in boolean_html_attributes:
                    if value:
                        value = name
                    else:
                        value = None

                if value is None:
                    try:
                        del attrs[name]
                    except KeyError:
                        pass
                else:
                    if type(value) not in (StringType, UnicodeType):
                        value = str(value)
                    attrs[name] = value

        for key, value in attrs.items():
            head += ' %s="%s"' % (key, value)
        head += '>'

        # Process the content
        content = ''
        if self.scontent is not None:
            expression = self.scontent

            value = expression.evaluate(stack, repeat)
            content = content + value
        else:
            for child in self.children:
                content = content + child.process(stack, repeat)

        # Remove the element but preserves its children if it is a stl:block
        if self.name == 'stl:block':
            return content

        foot = '</%s>' % self.name
        return head + content + foot


    def toxml(self):
        # Open tag
        s = '<' + self.name
        for key, value in self.attrs.items():
            s += ' %s="%s"' % (key, value)
        s += '>'

        # Content
        for child in self.children:
            s += child.toxml()

        # Close tag
        return s + '</' + self.name + '>'



class Document:
    def __init__(self):
        self.children = []


    def process(self, stack, repeat):
        s = ''
        for child in self.children:
            s = s + child.process(stack, repeat)
        return s


class Template:
    """ """

    def __init__(self, data, encoding='utf-8'):
        """
        The prefix parameter
        """
        if type(data) is UnicodeType:
            encoding = 'utf-8'
            data = data.encode(encoding)
        elif type(data) is not StringType:
            data = data.read()

        # Escape internal entities
        data = data.replace('&', '&amp;')

        file = StringIO(data)

        # Add parsing semantics
        p = expat.ParserCreate()
        p.StartElementHandler = self.start_element
        p.EndElementHandler = self.end_element
        p.DefaultHandler = self.default

        # Parse file
        document = Document()
        self.stack = [document]
        p.ParseFile(file)
        del self.stack

        self.document = document


    def start_element(self, name, attrs):
        dummy = name != 'stl:block' \
                and not [ x for x in attrs.keys() if x.startswith('stl:') ]

        element = Element(name, attrs)

        # Mark as dummy if needed
        if dummy:
            element.dummy = 1

        self.stack.append(element)


    def end_element(self, name):
        element = self.stack.pop()

        if getattr(element, 'dummy', 0) and not [ x for x in element.children
                                                  if isinstance(x, Element) ]:
            self.default(element.toxml())
        else:
            self.stack[-1].children.append(element)


    def default(self, data):
        # Unscape internal entities
        data = data.replace('&amp;', '&')

        children = element = self.stack[-1].children
        if children and isinstance(children[-1], Raw):
            children[-1].data += data
        else:
            children.append(Raw(data))


    def __call__(self, namespace={}):
        """ """
        stack = NamespaceStack()
        stack.append(namespace)

        repeat = NamespaceStack()

        return self.document.process(stack, repeat)



