# Copyright (C) 2002-2003 Juan David Ibez Palomar <jdavid@itaapy.com>
#               2003 Itaapy <contact@itaapy.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA


"""
Provides an extended DOM tree (based on minidom). The extended tree adds
methods that let to match nodes using XPath expressions and transform the
found nodes with Python functions.

These Python functions could be called rules, following the XSLT terminolgy.
Actually, this way to transform an XML document, using XPath expressions and
rules, is inspired by XSLT.

The extended tree methods start by the 'xpy' prefix because of historical
reasons, before this module was called 'xpy'.
"""

__revision__ = "$Id: tree.py,v 1.4 2003/08/05 09:11:26 jdavid Exp $"


# Python
from copy import deepcopy
from types import ListType, StringType, TupleType, UnicodeType

# PyXML
from xml.dom import minidom
from xml import xpath



class Node:
    """
    Represents a node of an XPY tree, adds the XPY API to the DOM tree
    nodes.
    """

    def xpy_apply(self, rule, expr='.', *args, **kw):
        """
        Applies a rule to all the nodes in the current context that match
        the expression. The parameters "args" and "kw" are passed to the
	rule (which is a function).
        """
        nodes = []
        for x in xpath.Evaluate(expr, self):
            x = apply(rule, (x,) + args, kw)
            nodes.append(x)
        return nodes


    # Manipulate the node
    def xpy_clone(self, deep):
        clone = self.cloneNode(deep)
        change_class(clone)
        return clone


    def xpy_search(self, expr):
        """
        Returns all the nodes that match the given XPath expression.
        """
        node_list = NodeList()
        for x in xpath.Evaluate(expr, self):
            node_list.append(x)
        return node_list


    def xpy_content(self, data):
        """
        Replaces the content of this node by the given data.
        """
        # Remove the current children
        self.childNodes = []

        if type(data) not in (ListType, TupleType):
            data = [data]

        for x in data:
            if type(x) in (StringType, UnicodeType):
                node = Text()
                node.data = x
                self.appendChild(node)
            else:
                x = deepcopy(x)

                # We suppose it is an dom tree, we should check it
                self.appendChild(x.childNodes[0])



if list is type([]):
    # Python 2.2 or later
    class NodeList(minidom.NodeList):
        pass
else:
    from UserList import UserList
    class NodeList(UserList):
        pass


def xpy_apply(self, rule, *args, **kw):
    """
    Applies a rule to all the nodes of this list. The parameters "args" and
    "kw" are passed to the rule (which is a function).
    """
    for x in self:
        apply(rule, (x,) + args, kw)


NodeList.xpy_apply = xpy_apply


class Element(Node, minidom.Element):
    def writexml(self, writer, indent="", addindent="", newl=""):
        """
        Overrides the default implementation to output empty elements
        as <foo></foo> instead of <foo/>, this is important because
        most browsers don't understand the abbreviated form.
        """
        # indent = current indentation
        # addindent = indentation to add to higher levels
        # newl = newline string
        writer.write(indent + "<" + self.tagName)

        attrs = self._get_attributes()
        a_names = attrs.keys()
        a_names.sort()

        for a_name in a_names:
            writer.write(' %s="' % a_name)
            minidom._write_data(writer, attrs[a_name].value)
            writer.write('"')
        if self.childNodes:
            writer.write(">%s" % (newl))
            for node in self.childNodes:
                node.writexml(writer, indent+addindent, addindent, newl)
            writer.write("%s</%s>%s" % (indent, self.tagName, newl))
        else:
            writer.write("></%s>%s" % (self.tagName, newl))


    def getcontent(self):
        """
        Returns an string with the node content.
        """
        # indent = the indentation string to prepend, per level
        # newl = the newline string to append
        writer = minidom._get_StringIO()
##        if encoding is not None:
##            import codecs
##            # Can't use codecs.getwriter to preserve 2.0 compatibility
##            writer = codecs.lookup(encoding)[3](writer)

        for node in self.childNodes:
            node.writexml(writer)

        return writer.getvalue()



class Attr(Node, minidom.Attr):
    pass


class Text(Node, minidom.Text):
    def writexml(self, writer, indent="", addindent="", newl=""):
        """
        We provide an alternative implementation, to prevent the scape
        of special characters.
        """
        data = "%s%s%s" % (indent, self.data, newl)
        writer.write(data)
        # To scape or not to scape, should be an option (XXX)
#        _write_data(writer, data) 


class CDATASection(Node, minidom.CDATASection):
    pass


class ProcessingInstruction(Node, minidom.ProcessingInstruction):
    pass


class Comment(Node, minidom.Comment):
    pass


class Document(Node, minidom.Document):
    pass


class DocumentType(Node, minidom.DocumentType):
    pass


class DocumentFragment(Node, minidom.DocumentFragment):
    pass



############################################################################
# Extend the tree.
############################################################################
classes = {minidom.Node.ELEMENT_NODE: Element,
           minidom.Node.ATTRIBUTE_NODE: Attr,
           minidom.Node.TEXT_NODE: Text,
           minidom.Node.CDATA_SECTION_NODE: CDATASection,
           minidom.Node.ENTITY_REFERENCE_NODE: None,
           minidom.Node.ENTITY_NODE: None,
           minidom.Node.PROCESSING_INSTRUCTION_NODE: ProcessingInstruction,
           minidom.Node.COMMENT_NODE: Comment,
           minidom.Node.DOCUMENT_NODE: Document,
           minidom.Node.DOCUMENT_TYPE_NODE: DocumentType,
           minidom.Node.DOCUMENT_FRAGMENT_NODE: DocumentFragment,
           minidom.Node.NOTATION_NODE: None}

def change_class(node):
    """
    Transforms the minidom tree to an extended tree. Does it changing the
    base class of all the nodes.
    """
    nodes = [node]
    while nodes:
        node = nodes.pop()
        node.__class__ = classes[node.nodeType]
        nodes.extend(node.childNodes)

    return node



############################################################################
# Parsers.
############################################################################

def parse_xml(data, encoding='utf-8'):
    """
    Parses an XML string and returns the extended DOM tree.
    """
    # Convert to unicode
    if type(data) is not UnicodeType:
        data = unicode(data, encoding)

    # Escape internal entities
    data = data.replace(u'&', u'&amp;')

    # Convert to UTF-8
    data = data.encode('utf-8')

    # Parse the data and build a normalized DOM
    dom = minidom.parseString(data)
    dom.normalize()

    # Change the classes the nodes are instance of
    change_class(dom)

    # Return the dom tree
    return dom


def parse_xmlfile(path, encoding='utf-8'):
    """
    Parses an XML file and returns the extended DOM tree.
    """
    data = open(path).read()
    return parse_xml(data, encoding)


def parse_text(data, encoding='utf-8'):
    """
    Parses a text string (not valid XML) and returns the extended DOM tree.
    """
    # Start with the XML declaration
    source = '<?xml version="1.0"?>\n'

    # Wrap the data within an div element
    source = source + '<div>%s</div>' % data
    return parse_xml(source, encoding)


def parse_textfile(path, encoding='utf-8'):
    """
    Parses a text file (not valid XML) and returns the extended DOM tree.
    """
    data = open(path).read()
    return parse_text(data, encoding)
